package org.objectweb.jonas.tools.migration.jboss;

import java.util.*;
import javax.xml.parsers.*;
import org.w3c.dom.*;

/**
 * Transformer
 *
 * @author Rafael H. Schloming &lt;rhs@mit.edu&gt;
 **/

public class Transformer {

    protected interface NodeFilter {
        boolean accept(Node node);
    }

    protected static final NodeFilter ALL = new NodeFilter() {
        public boolean accept(Node node) { return true; }
    };

    protected static final String LINE =
        System.getProperty("line.separator", "\n\r");

    private Stack m_stack = new Stack();
    private Document m_doc;
    private Node m_node;

    protected Transformer() {
        try {
            DocumentBuilder db = DocumentBuilderFactory.newInstance().newDocumentBuilder();
            m_doc = db.newDocument();
        } catch (ParserConfigurationException e) {
            throw new RuntimeException(e);
        }
        m_node = m_doc;
    }

    protected Document getDocument() {
        return m_doc;
    }

    private String indent(int level) {
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < level; i++) {
            buf.append("  ");
        }
        return buf.toString();
    }

    private void indent(boolean indent, int offset) {
        int level = m_stack.size() + offset;
        if (indent && level > 0) {
            text(LINE + indent(level));
        }
    }

    protected void open(String name) {
        open(name, true);
    }

    protected void open(String name, boolean indent) {
        indent(indent, 0);
        Element el = m_doc.createElement(name);
        m_node.appendChild(el);
        m_stack.push(m_node);
        m_node = el;
    }

    protected void close() {
        close(true);
    }

    protected void close(boolean indent) {
        indent(indent, -1);
        m_node = (Node) m_stack.pop();
    }

    protected void set(String name, String value) {
        if (m_node.getNodeType() != Node.ELEMENT_NODE) {
            throw new IllegalStateException
                ("not currently in an element");
        }
        Element el = (Element) m_node;
        el.setAttribute(name, value);
    }

    protected void text(String text) {
        m_node.appendChild(m_doc.createTextNode(text));
    }

    protected void comment(String comment) {
        m_node.appendChild(m_doc.createComment(comment));
    }

    protected void tag(String name, String value) {
        if (value == null) { return; }
        open(name);
        text(value);
        close(false);
    }

    protected void get(Node node, String name, Collection result,
                     NodeFilter filter) {
        NodeList nodes = node.getChildNodes();
        for (int i = 0; i < nodes.getLength(); i++) {
            Node child = nodes.item(i);
            if (child.getNodeName().equals(name) && filter.accept(child)) {
                result.add(child);
            }
        }
    }

    protected void query(Node node, String path, Collection result) {
        query(node, path, result, ALL);
    }

    protected void query(Node node, String path, Collection result,
                       NodeFilter filter) {
        int idx = path.indexOf('/');
        if (idx < 0) {
            get(node, path, result, filter);
            return;
        }

        String first = path.substring(0, idx);
        String rest = path.substring(idx+1);
        ArrayList l = new ArrayList();
        get(node, first, l, ALL);
        for (int i = 0; i < l.size(); i++) {
            Node n = (Node) l.get(i);
            query(n, rest, result, filter);
        }
    }

    protected List nodes(Node node, String path) {
        List result = new ArrayList();
        query(node, path, result);
        return result;
    }

    protected Node node(Node node, String path) {
        ArrayList result = new ArrayList();
        query(node, path, result);
        return singleton(result);
    }

    protected String value(Node node, String path) {
        if (node == null) { return null; }
        Node val = node(node, path + "/#text");
        if (val == null) {
            return null;
        } else {
            String result = val.getNodeValue();
            if (result == null) {
                throw new IllegalStateException
                    ("node has no value: " + val);
            } else {
                return result.trim();
            }
        }
    }

    protected List values(Node node, String path) {
        List result = new ArrayList();
        List nodes = nodes(node, path + "/#text");
        for (int i = 0; i < nodes.size(); i++) {
            Node nd = (Node) nodes.get(i);
            result.add(nd.getNodeValue());
        }
        return result;
    }

    protected Node singleton(Collection c) {
        if (c.isEmpty()) {
            return null;
        } else if (c.size() > 1) {
            throw new IllegalStateException
                ("too many results: " + c);
        } else {
            return (Node) c.iterator().next();
        }
    }

    private boolean isInlinable(NodeList nodes) {
        for (int i = 0; i < nodes.getLength(); i++) {
            Node node = nodes.item(i);
            switch (node.getNodeType()) {
            case Node.TEXT_NODE:
            case Node.ENTITY_REFERENCE_NODE:
                continue;
            default:
                return false;
            }
        }
        return true;
    }

    protected boolean isEmpty(String str) {
        return str == null || str.equals("");
    }

    protected void rename(Collection nodes, Map substitutions) {
        for (Iterator it = nodes.iterator(); it.hasNext(); ) {
            Node node = (Node) it.next();
            rename(node, substitutions);
        }
    }

    protected void rename(Node node, Map substitutions) {
        if (node == null) { return; }
        if (node.getNodeType() == Node.TEXT_NODE) {
            String value = node.getNodeValue().trim();
            if (!value.equals("")) {
                text(node.getNodeValue());
            }
            return;
        }
        String name = node.getNodeName();
        if (!substitutions.containsKey(name)) {
            return;
        }
        name = (String) substitutions.get(name);
        NodeList nodes = node.getChildNodes();
        boolean opened = false;
        if (!isEmpty(name)) {
            open(name);
            opened = true;
        }
        for (int i = 0; i < nodes.getLength(); i++) {
            rename(nodes.item(i), substitutions);
        }
        if (opened) {
            close(!isInlinable(nodes));
        }
    }

    protected static class Mapper extends LinkedHashMap {
        Mapper copy(String element) {
            return rename(element, element);
        }
        Mapper rename(String from, String to) {
            put(from, to);
            return this;
        }
        Mapper remove(String element) {
            return rename(element, "");
        }
    }

}
